local BaikCore = LibStub("AceAddon-3.0"):GetAddon("BaikCore")
local Assert = BaikCore.Assert
local CombatSafe = BaikCore.CombatSafe
local ChatCommand = BaikCore.ChatCommand
local Database = BaikCore.Database
local Option = BaikCore.Option
local OrderNumber = BaikCore.OrderNumber
local Reload = BaikCore.Reload
local Table = BaikCore.Table
local Ui = BaikCore.Ui

-- Add Module
local Grid = BaikCore:NewModule("Grid")
BaikCore.Grid = Grid

-- Constants
local GLOBAL_DB_ENABLED = "global_grid_on"
local GLOBAL_DB_COLOR = "global_grid_color"
local GLOBAL_DB_VERTICAL = "global_grid_vert"
local GLOBAL_DB_HORIZONTAL = "global_grid_horiz"
local GLOBAL_DB_THICKNESS = "global_grid_thickness"
local ENABLE_DEFAULT = false
local COLOR_DEFAULT = Ui:CreateColor(0, 0, 0, 0.75)
local VERT_DEFAULT = 32
local HORIZ_DEFAULT = 32
local THICKNESS_DEFAULT = 1
local NUMBER_PATTERN = "(-?%d+)"
local GRID_FRAME = nil

-- Private Variables
local orderNumber = OrderNumber:Create()

-- Private Methods
local function RoundUpPowTwo(number)
    local value = math.log(number) / math.log(2)
    local powTwo = math.ceil(value)

    return math.pow(2, powTwo)
end

local function ExtractNumber(valueString)
    local match = string.match(valueString, NUMBER_PATTERN)

    return tonumber(match)
end

local function CreateLine(parentFrame)
    local line = parentFrame:CreateTexture("BaikLine", "BACKGROUND")
    local color = Database:Get(GLOBAL_DB_COLOR)

    line:SetColorTexture(color.r, color.g, color.b, color.a)

    return line
end

local function GetGridFrame()
    if GRID_FRAME then
        return GRID_FRAME
    end

    local frame = CreateFrame("Frame", "BaikGrid", UIParent)

    -- Adjust to match full screen
    frame:SetAllPoints()

    -- Screen data
    local width = GetScreenWidth()
    local height = GetScreenHeight()
    local thickness = Database:Get(GLOBAL_DB_THICKNESS)

    -- Create horiz lines
    local numHoriz = Database:Get(GLOBAL_DB_HORIZONTAL)
    local incHoriz = height / numHoriz
    local horizLines = {}
    for yPos = 0, height, incHoriz do
        local line = CreateLine(frame)
        Ui:DrawLine(line, frame, 0, yPos, width, yPos, thickness)
        table.insert(horizLines, line)
    end
    frame.horizLines = horizLines

    -- Create vert lines
    local numVert = Database:Get(GLOBAL_DB_VERTICAL)
    local incVert = width / numVert
    local vertLines = {}
    for xPos = 0, width, incVert do
        local line = CreateLine(frame)
        Ui:DrawLine(line, frame, xPos, 0, xPos, height, thickness)
        table.insert(vertLines, line)
    end
    frame.vertLines = vertLines

    GRID_FRAME = frame

    return frame
end

local function SyncLocked()
    if Database:Get(GLOBAL_DB_ENABLED) then
        local gridFrame = GetGridFrame()
        gridFrame:Show()
    else
        -- Kind of a hack for lazy loading. We don't want to create grid frame
        -- if the user doesn't have it enabled
        if not GRID_FRAME then
            return
        end

        local gridFrame = GetGridFrame()
        gridFrame:Hide()
    end
end

local function Sync()
    if CombatSafe:InLockdown() then
        return
    end

    SyncLocked()
end

local function SyncColor(color)
    local gridFrame = GetGridFrame()
    local vertLines = gridFrame.vertLines
    for _, line in pairs(vertLines) do
        line:SetColorTexture(color.r, color.g, color.b, color.a)
    end

    local horizLines = gridFrame.horizLines
    for _, line in pairs(horizLines) do
        line:SetColorTexture(color.r, color.g, color.b, color.a)
    end
end

local function ToggleSync()
    local value = Database:Get(GLOBAL_DB_ENABLED)
    Database:Set(GLOBAL_DB_ENABLED, not value)

    Sync()
end

local function GetOpts()
    -- Helper methods
    function GridDisabled(...)
        return not Database:Get(GLOBAL_DB_ENABLED)
    end

    function GridEnabled(...)
        return Database:Get(GLOBAL_DB_ENABLED)
    end

    -- Options
    return {
        name = "Grid",
        type = "group",
        args = {
            enabled = {
                name = "Enable",
                type = "toggle",
                order = orderNumber:GetNext(),
                get = GridEnabled,
                set = function(self, value, ...)
                    Database:Set(GLOBAL_DB_ENABLED, value)

                    Sync()
                end
            },
            color = {
                name = "Color",
                type = "color",
                order = orderNumber:GetNext(),
                hasAlpha = true,
                disabled = GridDisabled,
                get = function(...)
                    local color = Database:Get(GLOBAL_DB_COLOR)
                    return color.r, color.g, color.b, color.a
                end,
                set = function(self, r, g, b, a, ...)
                    Assert:Number(r)
                    Assert:Number(g)
                    Assert:Number(b)
                    Assert:Number(a)
                    Database:Set(GLOBAL_DB_COLOR, Ui:CreateColor(r, g, b, a))
                    SyncColor()
                end
            },
            horiz = {
                name = "Horizontal Lines",
                type = "input",
                order = orderNumber:GetNext(),
                disabled = GridDisabled,
                usage = "Default (32) - must be power of two",
                pattern = NUMBER_PATTERN,
                get = function(...)
                    return tostring(Database:Get(GLOBAL_DB_HORIZONTAL))
                end,
                set = function(self, value, ...)
                    Assert:String(value)
                    local number = ExtractNumber(value)
                    local powTwo = RoundUpPowTwo(number)

                    Database:Set(GLOBAL_DB_HORIZONTAL, powTwo)
                    Reload:Enable()
                end
            },
            vert = {
                name = "Vertical Lines",
                type = "input",
                order = orderNumber:GetNext(),
                disabled = GridDisabled,
                usage = "Default (32) - must be power of two",
                pattern = NUMBER_PATTERN,
                get = function(...)
                    return tostring(Database:Get(GLOBAL_DB_VERTICAL))
                end,
                set = function(self, value, ...)
                    Assert:String(value)
                    local number = ExtractNumber(value)
                    local powTwo = RoundUpPowTwo(number)

                    Database:Set(GLOBAL_DB_VERTICAL, powTwo)
                    Reload:Enable()
                end
            },
            thick = {
                name = "Line Thickness",
                type = "range",
                disabled = GridDisabled,
                min = 1,
                max = 5,
                step = 1,
                get = function(...)
                    return Database:Get(GLOBAL_DB_THICKNESS)
                end,
                set = function(self, value, ...)
                    Assert:Number(value)

                    Database:Set(GLOBAL_DB_THICKNESS, value)
                    Reload:Enable()
                end,
            }
        }
    }
end

-- Ace3 Callbacks
function Grid:OnInitialize()
    Database:InitIfEmpty(GLOBAL_DB_ENABLED, ENABLE_DEFAULT)
    Database:InitIfEmpty(GLOBAL_DB_COLOR, COLOR_DEFAULT)
    Database:InitIfEmpty(GLOBAL_DB_VERTICAL, VERT_DEFAULT)
    Database:InitIfEmpty(GLOBAL_DB_HORIZONTAL, HORIZ_DEFAULT)
    Database:InitIfEmpty(GLOBAL_DB_THICKNESS, THICKNESS_DEFAULT)
    Option:AddOption(Option.NAME, "grid", GetOpts())
    CombatSafe:Setup(SyncLocked)
end

function Grid:OnEnable()
    ChatCommand:Register("grid", ToggleSync, "Toggles grid overlay")
    Sync()
end

function Grid:OnDisable()
end
